//////////////////////////////////////////////////////////////////
// mysql JDBC Driver
//
// Version : 0.9.2
//
// JDBC : 1.22
//
// LastModified: 12.03.1997
//
// Copyright (c) 1997 GWE Technologies Ltd. All rights reserved
// See http://gwe.co.uk/mysql/jdbc/copyright.html for full details

package gwe.sql;

import java.sql.*;
import gwe.sql.gweMysql;
import gwe.sql.gweMysqlException;
import java.util.ArrayList;
import java.util.Iterator;

/**
 * This class is the implementation of the JDBC draft specification
 * Statement interface.  Statement objects are used for executing SQL
 * statements.
 */
public class gweMysqlStatement implements java.sql.Statement {
  /**
   * A connection object to direct information to.
   */
  protected gweMysqlConnection connection;
  /**
   * A result set returned from this query or whatever.
   */
  private gweMysqlResultSetBuffer result;
  /**
   * The max field size for mysql
   * NOTE: This is the value of MAXBUFFER in gweMysql.packet
   */
  private int max_field_size = 65535;
  /**
   * The max rows supported by mysql - big!
   */
  private int max_rows = 50000000;   //Because it's in the FAQ!
  /**
   * The number of seconds the driver will allow for a SQL statement to
   * execute before giving up.  The default is to wait forever (0).
   */
  private int timeout = 0;
  /**
   * a hint as to the direction in which the rows in a result set
   */
  private int fetchDirection;
  /**
   * a hint as to the number of rows that should
   * be fetched from the database when more rows are needed.
   */
  private int fetchSize;
  /**
   * result set type
   */
  private int resultSetType;
  /**
   * result set concurrency
   */
  private int resultSetConcurrency;
  /**
   * batch of commmands
   */
  private java.util.ArrayList batch = null;

  /**
   * AutoGeneratedKey and Flag
   */
  private boolean  hasAutoIncrementValue;
  protected static final String defIncrementColName = "AUTO_INCREMENT";
  protected String autoIncrementColName;
  protected long   autoIncrementValue;

  /**
   * Constructs a new gweMysqlStatement object.
   * @param conn the gweMysqlConnection object
   * @param rs_type the result set type
   * @param rs_concur the result set concurrency
   */
  public gweMysqlStatement(gweMysqlConnection conn,
                           int rs_type, int rs_concur,
                           boolean hasAutoGeneratedKey) {
    connection = conn;
    resultSetType = rs_type;
    resultSetConcurrency = rs_concur;
    fetchDirection = ResultSet.FETCH_UNKNOWN;
    fetchSize = 0;

    hasAutoIncrementValue = hasAutoGeneratedKey;
    autoIncrementColName = defIncrementColName;
    autoIncrementValue = -1;
  }

  /**
   * JDBC draft specification method to execute the specified SQL
   * query and gives back a result set.
   * @see java.sql.Statement#executeQuery
   * @exception SQLException raised for any gweMysqlException
   * @param sql the SQL statement string
   * @return the result set from the query
   */
  public ResultSet executeQuery(String sql) throws SQLException {
    result = null;
    // to avoid creating new resultset before close old resultset!!
    connection.setCurResultSet(null);
    gweMysql r = connection.executeQuery(sql);
    gweMysqlResultSetCore curResultSet =
       new gweMysqlResultSetCore(r, this, resultSetType, resultSetConcurrency);
    connection.setCurResultSet(curResultSet);
    return new gweMysqlResultSetBuffer(curResultSet);
  }

  /**
   * JDBC 0.54 specification method for executing simple UPDATE, INSERT,
   * DELETE, etc. statements which do not return ResultSet's.
   * @see java.sql.Statement#executeUpdate
   * @exception java.sql.SQLException thrown when an error occurs executing
   * the SQL
   * @return 0
   */
  public int executeUpdate(String sql) throws SQLException {
    // to avoid send new SQL statement before close old resultset!!
    connection.setCurResultSet(null);
    long[] count_and_id = connection.executeUpdate(sql);
    autoIncrementValue = count_and_id[1];
    return (int)count_and_id[0];
  }

  /**
   * JDBC draft specification method to execute an SQL statement and return
   * true if a result set was gotten, false otherwise.
   * @see java.sql.Statement#execute
   * @exception SQLException raised for any gweMysqlException encountered
   * @param sql the SQL to be executed
   * @return true if there is a result set available
   */
  public boolean execute(String sql) throws SQLException {
    // to avoid creating new resultset before close old resultset!!
    connection.setCurResultSet(null);
    gweMysql r = connection.executeQuery(sql);
    if( r == null ) {
      result = null;
    }
    else {
      gweMysqlResultSetCore curResultSet =
       new gweMysqlResultSetCore(r, this, resultSetType, resultSetConcurrency);
      connection.setCurResultSet(curResultSet);
      result = new gweMysqlResultSetBuffer(curResultSet);
    }
    return (result != null);
  }

  /**
   * JDBC draft specification method to close off any open result sets
   * for this Statement.
   * @see java.sql.Statement#close
   */
  public void close() throws SQLException {
  }

  /**
   * JDBC 0.54 specification method for obtaining a result set from a
   * statement object.
   * @see java.sql.Statement#getResultSet
   * @return null if no result set is available, otherwise a result set
   */
  public ResultSet getResultSet() throws SQLException {
    ResultSet r;

    r = result;
    result = null;
    return r;
  }

  /**
   * Returns -1 always for mysql.
   * @see java.sql.Statement#getUpdateCount
   * @return -1
   */
  public int getUpdateCount() throws SQLException {
    return -1;
  }

  /**
   * JDBC 0.54 specification for determining if any more result sets
   * are left from a database statement.
   * @see java.sql.Statement#getMoreResults
   * @return true if rows are to be gotten
   */
  public boolean getMoreResults() throws SQLException {
    return (result != null);
  }

  /**
   * JDBC draft specification method for determining the number of
   * results sets left to process for multiple result set processing.
   * @see java.sql.Statement#getMoreCounts
   * @see getMoreResults
   * @return -1
   */
  public int getMoreCounts() throws SQLException {
    return -1;
  }

  /**
   * JDBC draft specification method for getting the max field size
   * @see java.sql.Statement#getMaxFieldSize
   * @return the value of max field size
   */
  public int getMaxFieldSize() throws SQLException {
    return max_field_size;
  }

  /**
   * JDBC draft specification method for setting the max field size.
   * @see java.sql.Statement#setMaxFieldSize
   * @param max the maximum field size
   */
  public void setMaxFieldSize(int max) throws SQLException {
    max_field_size = max;
  }

  /**
   * JDBC draft specification method for getting the max rows supported
   * @see java.sql.Statement#getMaxRows
   * @return the maximum rows supported by mysql
   */
  public int getMaxRows() throws SQLException {
    return max_rows;
  }

  /**
   * JDBC draft specification method for setting the max rows.
   * @see java.sql.Statement.setMaxRows
   * @param max the max rows
   */
  public void setMaxRows(int max) throws SQLException {
    max_rows = max;
  }

  /**
   * JDBC draft specification method for escape processing.
   * @see java.sql.Statement#setEscapeProcessing
   * @param enable this does nothing right now
   */
  public void setEscapeProcessing(boolean enable)
       throws SQLException {
    throw new SQLException("mysql does not support escape processing.");
  }

  /**
   * JDBC draft specification method for getting the query timeout, which
   * is the number of seconds the driver waits before giving up on the
   * query.
   * @see java.sql.Statement#getQueryTimeout
   * @see setQueryTimeout
   * @return the timeout value for this statement
   */
  public int getQueryTimeout() throws SQLException {
    return timeout;
  }

  /**
   * JDBC draft specification method for setting the query timeout.
   * @see java.sql.Statement#setQueryTimeout
   * @see getQueryTimeout
   * @param x the new query timeout value
   */
  public void setQueryTimeout(int x) throws SQLException {
    timeout = x;
  }

  /**
   * JDBC draft specification method for allowing one thread to cancel this
   * statement which may be running in another thread.
   * @see java.sql.Statement#cancel
   */
  public void cancel() {
  }

  /**
   * JDBC draft specification for getting the chain of warnings for this
   * statement.
   * @see java.sql.Statement#getWarnings
   * @return the chain of warnings
   */
  public final SQLWarning getWarnings() throws SQLException {
    return null;
  }

  /**
   * JDBC draft specification for clearing the warning chain.
   * @see java.sql.Statement#clearWarnings
   */
  public void clearWarnings() throws SQLException {
  }

  /**
   * JDBC 0.65 specification for setting cursor names.
   * mysql does not support cursors
   */
  public void setCursorName(String unused) throws SQLException {
    throw new SQLException("mysql does not support cursors.");
  }

  //--------------------------JDBC 2.0-----------------------------------

  public void setFetchDirection(int direction) throws SQLException {
    fetchDirection = direction;
  }

  public int getFetchDirection() throws SQLException {
    return fetchDirection;
  }

  public void setFetchSize(int rows) throws SQLException {
    fetchSize = rows;
  }

  public int getFetchSize() throws SQLException {
    return fetchSize;
  }

  public int getResultSetConcurrency() throws SQLException {
    return resultSetConcurrency;
  }

  public int getResultSetType() throws SQLException {
    return resultSetType;
  }

  private void ensureBatch() {
    if( batch == null )
      batch = new java.util.ArrayList();
  }

  public void addBatch(String sql) throws SQLException {
    ensureBatch();
    batch.add(sql);
  }

  public void clearBatch() throws SQLException {
    ensureBatch();
    batch.clear();
  }

  public int[] executeBatch() throws SQLException {
    ensureBatch();
    int[] retval = new int[batch.size()];

    for(int i = 0; i < retval.length; i++)
    {
      retval[i] = 0;
    }

    int j = 0;
    for(Iterator itr = batch.iterator(); itr.hasNext(); j++)
    {
      String sql = (String)itr.next();
      try{
        retval[j] = executeUpdate(sql);
      }
      catch(SQLException sql_e){
        new BatchUpdateException(retval);
      }
    }

    return (retval);
  }

  public Connection getConnection() throws SQLException {
    return connection;
  }

  //--------------------------JDBC 3.0-----------------------------

  boolean hasAutoGeneratedKey() {
    return hasAutoIncrementValue;
  }

  /**
   * Moves to this <code>Statement</code> object's next result, deals with
   * any current <code>ResultSet</code> object(s) according  to the instructions
   * specified by the given flag, and returns
   * <code>true</code> if the next result is a <code>ResultSet</code> object.
   *
   * <P>There are no more results when the following is true:
   * <PRE>
   *      <code>(!getMoreResults() && (getUpdateCount() == -1)</code>
   * </PRE>
   *
   * @param current one of the following <code>Statement</code>
   *        constants indicating what should happen to current
   *        <code>ResultSet</code> objects obtained using the method
   *        <code>getResultSet</code>:
   *        <code>CLOSE_CURRENT_RESULT</code>,
   *        <code>KEEP_CURRENT_RESULT</code>, or
   *        <code>CLOSE_ALL_RESULTS</code>
   * @return <code>true</code> if the next result is a <code>ResultSet</code>
   *         object; <code>false</code> if it is an update count or there are no
   *         more results
   * @exception SQLException if a database access error occurs
   * @since 1.4
   * @see #execute
   */
  public boolean getMoreResults(int current) throws SQLException {
    throw new SQLException("mysql does not support getMoreResults(int).");
  }
  /**
   * Retrieves any auto-generated keys created as a result of executing this
   * <code>Statement</code> object. If this <code>Statement</code> object did
   * not generate any keys, an empty <code>ResultSet</code>
   * object is returned.
   *
   * @return a <code>ResultSet</code> object containing the auto-generated key(s)
   *         generated by the execution of this <code>Statement</code> object
   * @exception SQLException if a database access error occurs
   * @since 1.4
   */
  public ResultSet getGeneratedKeys() throws SQLException {
    gweMysqlColumn[] cols = new gweMysqlColumn[1];
    cols[0] = new gweMysqlColumn("", autoIncrementColName, -1,
                           gweMysqlMetaResult.FIELD_TYPE_LONG, (short)-1, -1);

    ArrayList rows = new ArrayList();
    if( hasAutoGeneratedKey() )
    {
      ArrayList rowdata = new ArrayList();
      rowdata.add(new Long(autoIncrementValue)); // Key
      rows.add(rowdata);
    }
    return new gweMysqlVectorResultSet(connection, cols, rows);
  }
  /**
   * Executes the given SQL statement and signals the driver with the
   * given flag about whether the
   * auto-generated keys produced by this <code>Statement</code> object
   * should be made available for retrieval.
   *
   * @param sql must be an SQL <code>INSERT</code>, <code>UPDATE</code> or
   *        <code>DELETE</code> statement or an SQL statement that
   *        returns nothing
   * @param autoGeneratedKeys a flag indicating whether auto-generated keys
   *        should be made available for retrieval;
   *         one of the following constants:
   *         <code>Statement.RETURN_GENERATED_KEYS</code>
   *         <code>Statement.NO_GENERATED_KEYS</code>
   * @return either the row count for <code>INSERT</code>, <code>UPDATE</code>
   *         or <code>DELETE</code> statements, or <code>0</code> for SQL
   *         statements that return nothing
   * @exception SQLException if a database access error occurs, the given
   *            SQL statement returns a <code>ResultSet</code> object, or
   *            the given constant is not one of those allowed
   * @since 1.4
   */
  public int executeUpdate(String sql, int autoGeneratedKeys) throws SQLException {
    autoIncrementColName = defIncrementColName;
    if( autoGeneratedKeys == Statement.RETURN_GENERATED_KEYS )
      hasAutoIncrementValue = true;
    else if( autoGeneratedKeys == Statement.NO_GENERATED_KEYS )
      hasAutoIncrementValue = false;
    else
      throw new SQLException("Illegal argument at autoGeneratedKeys : executeUpdate(String, int).");

    return executeUpdate(sql);
  }
  /**
   * Executes the given SQL statement and signals the driver that the
   * auto-generated keys indicated in the given array should be made available
   * for retrieval.  The driver will ignore the array if the SQL statement
   * is not an <code>INSERT</code> statement.
   *
   * @param sql an SQL <code>INSERT</code>, <code>UPDATE</code> or
   *        <code>DELETE</code> statement or an SQL statement that returns nothing,
   *        such as an SQL DDL statement
   * @param columnIndexes an array of column indexes indicating the columns
   *        that should be returned from the inserted row
   * @return either the row count for <code>INSERT</code>, <code>UPDATE</code>,
   *         or <code>DELETE</code> statements, or 0 for SQL statements
   *         that return nothing
   * @exception SQLException if a database access error occurs or the SQL
   *            statement returns a <code>ResultSet</code> object
   * @since 1.4
   */
  public int executeUpdate(String sql, int[] columnIndexes) throws SQLException {
    if( columnIndexes.length == 1 && columnIndexes[0] == 1 )
    {
      autoIncrementColName = defIncrementColName;
      hasAutoIncrementValue = true;
      return executeUpdate(sql);
    }
    throw new SQLException("Illegal argument : executeUpdate(String, int[]).");
  }
  /**
   * Executes the given SQL statement and signals the driver that the
   * auto-generated keys indicated in the given array should be made available
   * for retrieval.  The driver will ignore the array if the SQL statement
   * is not an <code>INSERT</code> statement.
   *
   * @param sql an SQL <code>INSERT</code>, <code>UPDATE</code> or
   *        <code>DELETE</code> statement or an SQL statement that returns nothing
   * @param columnNames an array of the names of the columns that should be
   *        returned from the inserted row
   * @return either the row count for <code>INSERT</code>, <code>UPDATE</code>,
   *         or <code>DELETE</code> statements, or 0 for SQL statements
   *         that return nothing
   * @exception SQLException if a database access error occurs
   *
   * @since 1.4
   */
  public int executeUpdate(String sql, String[] columnNames) throws SQLException {
    if( columnNames.length == 1 && columnNames[0] != null )
    {
      autoIncrementColName = columnNames[0];
      hasAutoIncrementValue = true;
      return executeUpdate(sql);
    }
    throw new SQLException("Illegal argument : executeUpdate(String, String[]).");
  }
  /**
   * Executes the given SQL statement, which may return multiple results,
   * and signals the driver that any
   * auto-generated keys should be made available
   * for retrieval.  The driver will ignore this signal if the SQL statement
   * is not an <code>INSERT</code> statement.
   * <P>
   * In some (uncommon) situations, a single SQL statement may return
   * multiple result sets and/or update counts.  Normally you can ignore
   * this unless you are (1) executing a stored procedure that you know may
   * return multiple results or (2) you are dynamically executing an
   * unknown SQL string.
   * <P>
   * The <code>execute</code> method executes an SQL statement and indicates the
   * form of the first result.  You must then use the methods
   * <code>getResultSet</code> or <code>getUpdateCount</code>
   * to retrieve the result, and <code>getMoreResults</code> to
   * move to any subsequent result(s).
   *
   * @param sql any SQL statement
   * @param autoGeneratedKeys a constant indicating whether auto-generated
   *        keys should be made available for retrieval using the method
   *        <code>getGeneratedKeys</code>; one of the following constants:
   *        <code>Statement.RETURN_GENERATED_KEYS</code> or
   *	      <code>Statement.NO_GENERATED_KEYS</code>
   * @return <code>true</code> if the first result is a <code>ResultSet</code>
   *         object; <code>false</code> if it is an update count or there are
   *         no results
   * @exception SQLException if a database access error occurs
   * @see #getResultSet
   * @see #getUpdateCount
   * @see #getMoreResults
   * @see #getGeneratedKeys
   *
   * @since 1.4
   */
  public boolean execute(String sql, int autoGeneratedKeys) throws SQLException {
    autoIncrementColName = defIncrementColName;
    if( autoGeneratedKeys == Statement.RETURN_GENERATED_KEYS )
    {
      hasAutoIncrementValue = true;
      executeUpdate(sql);
      return false;
    }
    else
    {
      hasAutoIncrementValue = false;
      return execute(sql);
    }
  }
  /**
   * Executes the given SQL statement, which may return multiple results,
   * and signals the driver that the
   * auto-generated keys indicated in the given array should be made available
   * for retrieval.  This array contains the indexes of the columns in the
   * target table that contain the auto-generated keys that should be made
   * available. The driver will ignore the array if the given SQL statement
   * is not an <code>INSERT</code> statement.
   * <P>
   * Under some (uncommon) situations, a single SQL statement may return
   * multiple result sets and/or update counts.  Normally you can ignore
   * this unless you are (1) executing a stored procedure that you know may
   * return multiple results or (2) you are dynamically executing an
   * unknown SQL string.
   * <P>
   * The <code>execute</code> method executes an SQL statement and indicates the
   * form of the first result.  You must then use the methods
   * <code>getResultSet</code> or <code>getUpdateCount</code>
   * to retrieve the result, and <code>getMoreResults</code> to
   * move to any subsequent result(s).
   *
   * @param sql any SQL statement
   * @param columnIndexes an array of the indexes of the columns in the
   *        inserted row that should be  made available for retrieval by a
   *        call to the method <code>getGeneratedKeys</code>
   * @return <code>true</code> if the first result is a <code>ResultSet</code>
   *         object; <code>false</code> if it is an update count or there
   *         are no results
   * @exception SQLException if a database access error occurs
   * @see #getResultSet
   * @see #getUpdateCount
   * @see #getMoreResults
   *
   * @since 1.4
   */
  public boolean execute(String sql, int[] columnIndexes) throws SQLException {
    if( columnIndexes.length == 1 && columnIndexes[0] == 1 )
    {
      autoIncrementColName = defIncrementColName;
      hasAutoIncrementValue = true;
      return execute(sql);
    }
    throw new SQLException("Illegal argument : execute(String, int[]).");
  }
  /**
   * Executes the given SQL statement, which may return multiple results,
   * and signals the driver that the
   * auto-generated keys indicated in the given array should be made available
   * for retrieval. This array contains the names of the columns in the
   * target table that contain the auto-generated keys that should be made
   * available. The driver will ignore the array if the given SQL statement
   * is not an <code>INSERT</code> statement.
   * <P>
   * In some (uncommon) situations, a single SQL statement may return
   * multiple result sets and/or update counts.  Normally you can ignore
   * this unless you are (1) executing a stored procedure that you know may
   * return multiple results or (2) you are dynamically executing an
   * unknown SQL string.
   * <P>
   * The <code>execute</code> method executes an SQL statement and indicates the
   * form of the first result.  You must then use the methods
   * <code>getResultSet</code> or <code>getUpdateCount</code>
   * to retrieve the result, and <code>getMoreResults</code> to
   * move to any subsequent result(s).
   *
   * @param sql any SQL statement
   * @param columnNames an array of the names of the columns in the inserted
   *        row that should be made available for retrieval by a call to the
   *        method <code>getGeneratedKeys</code>
   * @return <code>true</code> if the next result is a <code>ResultSet</code>
   *         object; <code>false</code> if it is an update count or there
   *         are no more results
   * @exception SQLException if a database access error occurs
   * @see #getResultSet
   * @see #getUpdateCount
   * @see #getMoreResults
   * @see #getGeneratedKeys
   *
   * @since 1.4
   */
  public boolean execute(String sql, String[] columnNames) throws SQLException {
    if( columnNames.length == 1 && columnNames[0] != null )
    {
      autoIncrementColName = columnNames[0];
      hasAutoIncrementValue = true;
      return execute(sql);
    }
    throw new SQLException("Illegal argument : execute(String, String[]).");
  }
  /**
   * Retrieves the result set holdability for <code>ResultSet</code> objects
   * generated by this <code>Statement</code> object.
   *
   * @return either <code>ResultSet.HOLD_CURSORS_OVER_COMMIT</code> or
   *         <code>ResultSet.CLOSE_CURSORS_AT_COMMIT</code>
   * @exception SQLException if a database access error occurs
   *
   * @since 1.4
   */
  public int getResultSetHoldability() throws SQLException {
    throw new SQLException("mysql does not support getResultSetHoldability().");
  }
}
